@@ -13,7 +13,7 @@
 #include <string.h>
 #include <stdint.h>
 
-#include "list.h"
+#include "util.h"
 #include "treeview.h"
 
 #ifndef MAX_DEPTH
@@ -28,7 +28,7 @@ struct vdi_tree {
 	char label[256];
 	uint32_t vid;
 	uint32_t pvid;
-	int highlight;
+	bool highlight;
 	struct list_head children;
 	struct list_head siblings;
 };
@@ -53,17 +53,17 @@ static struct vdi_tree *find_vdi(struct vdi_tree *parent, uint32_t vid,
 }
 
 static struct vdi_tree *new_vdi(const char *name, const char *label,
-				uint64_t vid, uint64_t pvid, int highlight)
+				uint64_t vid, uint64_t pvid, bool highlight)
 {
 	struct vdi_tree *vdi;
 
 	vdi = malloc(sizeof(struct vdi_tree));
 	if (!vdi) {
-		fprintf(stderr, "malloc\n");
+		fprintf(stderr, "Failed to allocate memory\n");
 		return NULL;
 	}
-	strcpy(vdi->name, name);
-	strcpy(vdi->label, label);
+	pstrcpy(vdi->name, sizeof(vdi->name), name);
+	pstrcpy(vdi->label, sizeof(vdi->label), label);
 	vdi->vid = vid;
 	vdi->pvid = pvid;
 	vdi->highlight = highlight;
@@ -77,7 +77,7 @@ void init_tree(void)
 }
 
 void add_vdi_tree(const char *name, const char *label, uint32_t vid,
-		  uint32_t pvid, int highlight)
+		  uint32_t pvid, bool highlight)
 {
 	struct vdi_tree *vdi, *parent;
 
@@ -98,10 +98,8 @@ static void compaction(struct vdi_tree *parent)
 
 	list_for_each_entry_safe(vdi, e, &parent->children, siblings) {
 		new_parent = find_vdi(root, vdi->pvid, vdi->name);
-		if (new_parent && parent != new_parent) {
-			list_del(&vdi->siblings);
-			list_add_tail(&vdi->siblings, &new_parent->children);
-		}
+		if (new_parent && parent != new_parent)
+			list_move_tail(&vdi->siblings, &new_parent->children);
 
 		compaction(vdi);
 	}
@@ -126,7 +124,7 @@ static void spaces(int n)
 		putchar(' ');
 }
 
-static void indent(int level, int first, int last)
+static void indent(int level, bool first, bool last)
 {
 	int lvl;
 
@@ -143,7 +141,7 @@ static void indent(int level, int first, int last)
 	}
 }
 
-static void _dump_tree(struct vdi_tree *current, int level, int first, int last)
+static void _dump_tree(struct vdi_tree *current, int level, bool first, bool last)
 {
 	struct vdi_tree *vdi;
 
@@ -184,7 +182,7 @@ void dump_tree(void)
 	width = malloc(sizeof(int) * depth);
 	more = malloc(sizeof(int) * depth);
 	if (!width || !more) {
-		fprintf(stderr, "out of memory\n");
+		fprintf(stderr, "Failed to allocate memory\n");
 		return;
 	}
 
@@ -192,6 +190,6 @@ void dump_tree(void)
 		printf("%s", vdi->name);
 		more[0] = 0;
 		width[0] = strlen(vdi->name);
-		_dump_tree(vdi, 1, 1, 1);
+		_dump_tree(vdi, 1, true, true);
 	}
 }
