@@ -15,6 +15,7 @@
 #include <errno.h>
 #include <string.h>
 #include <inttypes.h>
+#include <stdbool.h>
 #include <pthread.h>
 #include <stdio.h>
 #include <unistd.h>
@@ -24,135 +25,92 @@
 #include <sys/types.h>
 #include <sys/eventfd.h>
 #include <linux/types.h>
+#include <urcu/uatomic.h>
 
 #include "list.h"
 #include "util.h"
 #include "work.h"
 #include "logger.h"
 #include "event.h"
+#include "trace/trace.h"
 
 static int efd;
-static LIST_HEAD(worker_info_list);
-
-struct work_queue {
-	int wq_state;
-	int nr_active;
-	struct list_head pending_list;
-	struct list_head blocked_list;
-};
+int total_ordered_workers;
+LIST_HEAD(worker_info_list);
 
 enum wq_state {
-	WQ_BLOCKED = (1U << 0),
 	WQ_DEAD = (1U << 1),
 };
 
-struct worker_info {
-	struct list_head worker_info_siblings;
-
-	int nr_threads;
-
-	pthread_mutex_t finished_lock;
-	struct list_head finished_list;
-
-	/* wokers sleep on this and signaled by tgtd */
-	pthread_cond_t pending_cond;
-	/* locked by tgtd and workers */
-	pthread_mutex_t pending_lock;
-	/* protected by pending_lock */
-	struct work_queue q;
-
-	pthread_mutex_t startup_lock;
-
-	pthread_t worker_thread[0];
+/*
+ * Short thread is created on demand and destroyed after serving the work for
+ * gateway or io requests, aiming to solve two problems:
+ *
+ *  1. timeout of IO requests from guests. With on-demand short threads, we
+ *     guarantee that there is always one thread available to execute the
+ *     request as soon as possible.
+ *  2. sheep halt for corner case that all gateway and io threads are executing
+ *     local requests that ask for creation of another thread to execute the
+ *     requests and sleep-wait for responses.
+ */
+struct short_work {
+	struct work *work;
+	struct worker_info *wi;
 };
 
-static void work_queue_set_blocked(struct work_queue *q)
+static void *run_short_thread(void *arg)
 {
-	q->wq_state |= WQ_BLOCKED;
+	struct short_work *sw = arg;
+	eventfd_t value = 1;
+	static unsigned long idx;
+	int err;
+
+	/* Tell runtime to release resources after termination */
+	err = pthread_detach(pthread_self());
+	if (err)
+		panic("%s\n", strerror(err));
+
+	set_thread_name(sw->wi->name, uatomic_add_return(&idx, 1));
+
+	sw->work->fn(sw->work);
+
+	pthread_mutex_lock(&sw->wi->finished_lock);
+	list_add_tail(&sw->work->w_list, &sw->wi->finished_list);
+	pthread_mutex_unlock(&sw->wi->finished_lock);
+
+	eventfd_write(efd, value);
+	free(sw);
+	pthread_exit(NULL);
 }
 
-static void work_queue_clear_blocked(struct work_queue *q)
+static inline void create_short_thread(struct worker_info *wi,
+				       struct work *work)
 {
-	q->wq_state &= ~WQ_BLOCKED;
-}
+	pthread_t thread;
+	struct short_work *sw = xmalloc(sizeof *sw);
+	int err;
 
-static int work_queue_blocked(struct work_queue *q)
-{
-	return q->wq_state & WQ_BLOCKED;
-}
+	sw->work = work;
+	sw->wi = wi;
 
-static int work_enabled(struct work_queue *q, struct work *w)
-{
-	int enabled = 0;
-
-	switch (w->attr) {
-	case WORK_SIMPLE:
-		if (!work_queue_blocked(q))
-			enabled = 1;
-		break;
-	case WORK_ORDERED:
-		if (!work_queue_blocked(q) && !q->nr_active)
-			enabled = 1;
-		break;
-	default:
-		enabled = -1;
-	}
-
-	return enabled;
-}
-
-static void work_post_queued(struct work_queue *q, struct work *w)
-{
-	q->nr_active++;
-	if (w->attr == WORK_ORDERED)
-		work_queue_set_blocked(q);
-}
-
-static void __queue_work(struct work_queue *q, struct work *work, int enabled)
-{
-	struct worker_info *wi = container_of(q, struct worker_info, q);
-
-	if (enabled) {
-		pthread_mutex_lock(&wi->pending_lock);
-
-		list_add_tail(&work->w_list, &wi->q.pending_list);
-
-		pthread_mutex_unlock(&wi->pending_lock);
-
-		pthread_cond_signal(&wi->pending_cond);
-
-		work_post_queued(q, work);
-	} else
-		list_add_tail(&work->w_list, &wi->q.blocked_list);
+	err = pthread_create(&thread, NULL, run_short_thread, sw);
+	if (err)
+		panic("%s\n", strerror(err));
+	short_thread_begin();
 }
 
 void queue_work(struct work_queue *q, struct work *work)
 {
-	int enabled;
+	struct worker_info *wi = container_of(q, struct worker_info, q);
 
-	if (!list_empty(&q->blocked_list))
-		enabled = 0;
-	else
-		enabled = work_enabled(q, work);
+	if (wi->ordered) {
+		pthread_mutex_lock(&wi->pending_lock);
+		list_add_tail(&work->w_list, &wi->q.pending_list);
+		pthread_mutex_unlock(&wi->pending_lock);
 
-	__queue_work(q, work, enabled);
-}
-
-static void work_post_done(struct work_queue *q, enum work_attr attr)
-{
-	struct work *n, *t;
-
-	q->nr_active--;
-	if (attr == WORK_ORDERED)
-		work_queue_clear_blocked(q);
-
-	list_for_each_entry_safe(n, t, &q->blocked_list, w_list) {
-		if (!work_enabled(q, n))
-			break;
-
-		list_del(&n->w_list);
-		__queue_work(q, n, 1);
-	}
+		pthread_cond_signal(&wi->pending_cond);
+	} else
+		create_short_thread(wi, work);
 }
 
 static void bs_thread_request_done(int fd, int events, void *data)
@@ -173,17 +131,12 @@ static void bs_thread_request_done(int fd, int events, void *data)
 		pthread_mutex_unlock(&wi->finished_lock);
 
 		while (!list_empty(&list)) {
-			enum work_attr attr;
 			work = list_first_entry(&list, struct work, w_list);
 			list_del(&work->w_list);
 
-			/*
-			 * work->done might free the work so we must
-			 * save its attr for qork_post_done().
-			 */
-			attr = work->attr;
-			work->done(work, 0);
-			work_post_done(&wi->q, attr);
+			work->done(work);
+			if (!wi->ordered)
+				short_thread_end();
 		}
 	}
 }
@@ -192,18 +145,12 @@ static void *worker_routine(void *arg)
 {
 	struct worker_info *wi = arg;
 	struct work *work;
-	int i, idx = 0;
 	eventfd_t value = 1;
 
-	for (i = 0; i < wi->nr_threads; i++) {
-		if (wi->worker_thread[i] == pthread_self()) {
-			idx = i;
-			break;
-		}
-	}
+	set_thread_name(wi->name, 0);
 
 	pthread_mutex_lock(&wi->startup_lock);
-	dprintf("started this thread %d\n", idx);
+	/* started this thread */
 	pthread_mutex_unlock(&wi->startup_lock);
 
 	while (!(wi->q.wq_state & WQ_DEAD)) {
@@ -225,7 +172,7 @@ retest:
 		list_del(&work->w_list);
 		pthread_mutex_unlock(&wi->pending_lock);
 
-		work->fn(work, idx);
+		work->fn(work);
 
 		pthread_mutex_lock(&wi->finished_lock);
 		list_add_tail(&work->w_list, &wi->finished_list);
@@ -237,63 +184,59 @@ retest:
 	pthread_exit(NULL);
 }
 
-static int init_eventfd(void)
+int init_wqueue_eventfd(void)
 {
 	int ret;
-	static int done = 0;
-
-	if (done++)
-		return 0;
 
 	efd = eventfd(0, EFD_NONBLOCK);
 	if (efd < 0) {
-		eprintf("failed to create an event fd, %m\n");
+		sd_eprintf("failed to create an event fd: %m\n");
 		return 1;
 	}
 
 	ret = register_event(efd, bs_thread_request_done, NULL);
+	if (ret) {
+		sd_eprintf("failed to register event fd %m\n");
+		close(efd);
+		return 1;
+	}
 
 	return 0;
 }
 
-struct work_queue *init_work_queue(int nr)
+struct work_queue *init_work_queue(const char *name, bool ordered)
 {
-	int i, ret;
+	int ret;
 	struct worker_info *wi;
 
-	ret = init_eventfd();
-	if (ret)
-		return NULL;
+	wi = xzalloc(sizeof(*wi));
+	wi->name = name;
+	wi->ordered = ordered;
 
-	wi = zalloc(sizeof(*wi) + nr * sizeof(pthread_t));
-	if (!wi)
-		return NULL;
-
-	wi->nr_threads = nr;
-
-	INIT_LIST_HEAD(&wi->q.pending_list);
-	INIT_LIST_HEAD(&wi->q.blocked_list);
 	INIT_LIST_HEAD(&wi->finished_list);
 
-	pthread_cond_init(&wi->pending_cond, NULL);
-
 	pthread_mutex_init(&wi->finished_lock, NULL);
-	pthread_mutex_init(&wi->pending_lock, NULL);
-	pthread_mutex_init(&wi->startup_lock, NULL);
 
-	pthread_mutex_lock(&wi->startup_lock);
-	for (i = 0; i < wi->nr_threads; i++) {
-		ret = pthread_create(&wi->worker_thread[i], NULL,
-				     worker_routine, wi);
+	if (ordered) {
+		INIT_LIST_HEAD(&wi->q.pending_list);
 
+		pthread_cond_init(&wi->pending_cond, NULL);
+		pthread_mutex_init(&wi->pending_lock, NULL);
+		pthread_mutex_init(&wi->startup_lock, NULL);
+
+		pthread_mutex_lock(&wi->startup_lock);
+
+		ret = pthread_create(&wi->worker_thread, NULL, worker_routine,
+				     wi);
 		if (ret) {
-			eprintf("failed to create a worker thread, %d %s\n",
-				i, strerror(ret));
-			if (ret)
-				goto destroy_threads;
+			sd_eprintf("failed to create worker thread: %s\n",
+				strerror(ret));
+			goto destroy_threads;
 		}
+
+		pthread_mutex_unlock(&wi->startup_lock);
+		total_ordered_workers++;
 	}
-	pthread_mutex_unlock(&wi->startup_lock);
 
 	list_add(&wi->worker_info_siblings, &worker_info_list);
 
@@ -302,10 +245,8 @@ destroy_threads:
 
 	wi->q.wq_state |= WQ_DEAD;
 	pthread_mutex_unlock(&wi->startup_lock);
-	for (; i > 0; i--) {
-		pthread_join(wi->worker_thread[i - 1], NULL);
-		eprintf("stopped the worker thread %d\n", i - 1);
-	}
+	pthread_join(wi->worker_thread, NULL);
+	sd_eprintf("stopped worker thread\n");
 
 /* destroy_cond_mutex: */
 	pthread_cond_destroy(&wi->pending_cond);
@@ -319,15 +260,12 @@ destroy_threads:
 #ifdef COMPILE_UNUSED_CODE
 static void exit_work_queue(struct work_queue *q)
 {
-	int i;
 	struct worker_info *wi = container_of(q, struct worker_info, q);
 
 	q->wq_state |= WQ_DEAD;
 	pthread_cond_broadcast(&wi->pending_cond);
 
-	for (i = 0; wi->worker_thread[i] &&
-		     i < wi->nr_threads; i++)
-		pthread_join(wi->worker_thread[i], NULL);
+	pthread_join(wi->worker_thread, NULL);
 
 	pthread_cond_destroy(&wi->pending_cond);
 	pthread_mutex_destroy(&wi->pending_lock);
